'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.installLibrariesByUrls = exports.checkLibrariesInstalledByUrls = exports.getLibraryNameFromUrl = undefined;

var _ramda = require('ramda');

var R = _interopRequireWildcard(_ramda);

var _fsExtra = require('fs-extra');

var fse = _interopRequireWildcard(_fsExtra);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _ramdaFantasy = require('ramda-fantasy');

var _xodFuncTools = require('xod-func-tools');

var _download = require('./download');

var _download2 = _interopRequireDefault(_download);

var _unzip = require('./unzip');

var _unzip2 = _interopRequireDefault(_unzip);

var _progress = require('./progress');

var _progress2 = _interopRequireDefault(_progress);

var _messages = require('./messages');

var _messages2 = _interopRequireDefault(_messages);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

// :: URL -> String
const getProjectNameFromGithubUrl = R.match(/github.com\/.+\/(.+)\/{0,1}/);

// :: URL -> URL
const getGithubArchiveUrl = R.compose(R.concat(R.__, '/archive/master.zip'), R.when(R.pipe(R.tail, R.equals('/')), R.init));

/**
 * Arduino-builder does not work with directory names containing symbols
 * other than A-Za-z0-9_.
 * So the library root directory have to be renamed:
 * - Replace all hyphens with underscores
 * - Remove all symbols not in A-Za-z0-9_
 *
 * :: String -> String
 */
const normalizeLibraryName = R.compose(R.replace(/[^A-Za-z0-9_]/g, ''), R.replace(/-/g, '_'));

// :: Path -> String -> Promise Boolean Error
const checkLibraryInstalled = R.curry((libsPath, libName) => R.compose(fse.pathExists, _path2.default.resolve)(libsPath, libName));

// =============================================================================
//
// API
//
// =============================================================================

/**
 * Returns the normalized library name from a URL wrapped with `Maybe`.
 * Works only with URLs on the master branch of the GitHub repository.
 * If the URL is incorrect or the repository name can't be normalized at all
 * (contains only forbidden characters) it returns Nothing.
 *
 * :: URL -> Maybe String
 */
const getLibraryNameFromUrl = exports.getLibraryNameFromUrl = R.compose(R.chain(R.compose(R.ifElse(R.isEmpty, _ramdaFantasy.Maybe.Nothing, _ramdaFantasy.Maybe.of), normalizeLibraryName)), _ramdaFantasy.Maybe, R.nth(1), getProjectNameFromGithubUrl);

/**
 * Checks whether libraries already installed in the specified path.
 * Accepts a function as the first argument to notify someone about progress
 * (see `ProgressData` type in the `progress.js`).
 *
 * Returns `Promise` with the map of check results indexed by library URLs.
 * Returns rejected `Promise` either on a normalization library name error or
 * on an FS permission error.
 *
 * Example of the resulting map:
 * {
 *   'https://github.com/arduino-libraries/Stepper': true, // means it's installed
 *   'https://github.com/arduino-libraries/GSM': false,    // means it's missing
 * }
 *
 * :: (ProgressData -> _) -> Path -> [URL] -> Promise (Map URL Boolean) Error
 */
const checkLibrariesInstalledByUrls = exports.checkLibrariesInstalledByUrls = R.curry((onProgress, libsPath, libUrls) => {
  const progress = (0, _progress2.default)(libUrls.length);
  return R.compose((0, _xodFuncTools.then)(R.zipObj(libUrls)), _xodFuncTools.allPromises, R.map(url => R.compose((0, _xodFuncTools.foldMaybeWith)(() => Promise.reject((0, _xodFuncTools.createError)('CANT_GET_LIBRARY_NAME', { url })), libName => checkLibraryInstalled(libsPath, libName).then(R.tap(isInstalled => onProgress(progress(isInstalled ? _messages2.default.LIBRARY_EXIST({ libName }).note // eslint-disable-line new-cap
  : _messages2.default.LIBRARY_MISSING({ libName }).note // eslint-disable-line new-cap
  ))))), getLibraryNameFromUrl)(url)))(libUrls);
});

/**
 * Installs libraries in the specified path from the list of URLs.
 * Accepts a function as the first argument to notify someone about progress
 * (see `ProgressData` type in the `progress.js`).
 *
 * Returns `Promise` with the list of normalized names of installed libraries.
 * Returns rejected `Promise` on a normalization library name error,
 * on an FS permission error or on a connection/server error.
 *
 * :: (ProgressData -> _) -> Path -> [URL] -> Promise [String] Error
 */
const installLibrariesByUrls = exports.installLibrariesByUrls = R.curry((onProgress, libsPath, libUrls) => {
  const startFrom = libUrls.length; // because `check` uses first 3 steps
  const steps = libUrls.length * 3; // check, download, unpack
  const progress = (0, _progress2.default)(steps, startFrom);
  const progressCheckInstalled = R.pipe(progress.wrap, onProgress);

  return R.composeP(_xodFuncTools.allPromises, R.map(urlToInstall => {
    const libName = R.pipe(getLibraryNameFromUrl, (0, _xodFuncTools.explodeMaybe)('IMPOSIBLE ERROR') // because `checkLibrariesInstalledByUrls` already checked all URLs
    )(urlToInstall);
    const archiveUrl = getGithubArchiveUrl(urlToInstall);
    const libraryDir = _path2.default.resolve(libsPath, libName);
    const archivePath = _path2.default.resolve(libsPath, `${libName}.zip`);
    return (0, _download2.default)(archiveUrl, archivePath).then(R.tap(() =>
    // eslint-disable-next-line new-cap
    onProgress(progress(_messages2.default.LIBRARY_DOWNLOADED({ libName }).note)))).then(_unzip2.default).then(unpackedDir => fse.rename(_path2.default.resolve(libsPath, unpackedDir), libraryDir)).then(() => fse.remove(archivePath)).then(() =>
    // eslint-disable-next-line new-cap
    onProgress(progress(_messages2.default.LIBRARY_INSTALLED({ libName }).note))).then(() => libName);
  }), R.keys, R.filter(R.not), checkLibrariesInstalledByUrls(progressCheckInstalled))(libsPath, libUrls);
});